hopscotch.o hopscotch.d : hopscotch.c os.h sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h align.h target-os.h darwin-os.h target-arch-os.h \
  arch-os-generic.inc target-arch.h gc-internal.h genesis/code.h \
  genesis/simple-fun.h thread.h globals.h gencgc-alloc-region.h gc.h \
  genesis/symbol.h genesis/static-symbols.h genesis/thread.h \
  genesis/fdefn.h genesis/vector.h interrupt.h validate.h interr.h \
  gc-assert.h code.h gencgc-internal.h hopscotch.h genesis/bignum.h \
  murmur_hash.h
