x86-64-arch.o x86-64-arch.d : x86-64-arch.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h align.h globals.h os.h target-os.h darwin-os.h \
  target-arch-os.h arch-os-generic.inc target-arch.h validate.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h genesis/vector.h interrupt.h arch.h \
  lispregs.h target-lispregs.h genesis/regnames.h alloc.h gc-internal.h \
  genesis/code.h genesis/simple-fun.h interr.h gc-assert.h code.h \
  gencgc-internal.h hopscotch.h genesis/bignum.h genesis/sap.h \
  breakpoint.h pseudo-atomic.h unaligned.h search.h forwarding-ptr.h \
  core.h
