x86-64-darwin-os.o x86-64-darwin-os.d : x86-64-darwin-os.c thread.h sbcl.h genesis/config.h \
  genesis/constants.h globals.h runtime.h fixnump.h align.h os.h \
  target-os.h darwin-os.h target-arch-os.h arch-os-generic.inc \
  target-arch.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h \
  genesis/vector.h interrupt.h validate.h x86-64-darwin-os.h \
  x86-64-arch.h gc-internal.h genesis/code.h genesis/simple-fun.h \
  interr.h gc-assert.h code.h gencgc-internal.h hopscotch.h \
  genesis/bignum.h arch.h
