x86-64-assem.o x86-64-assem.d : x86-64-assem.S genesis/config.h validate.h sbcl.h \
  genesis/constants.h genesis/closure.h genesis/fdefn.h \
  genesis/static-symbols.h genesis/thread.h genesis/ratio.h \
  genesis/complex.h genesis/vector.h
