dynbind.o dynbind.d : dynbind.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h align.h globals.h os.h target-os.h darwin-os.h \
  target-arch-os.h arch-os-generic.inc target-arch.h dynbind.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h genesis/vector.h interrupt.h \
  validate.h interr.h genesis/binding.h
