coreparse.o coreparse.d : coreparse.c sbcl.h genesis/config.h genesis/constants.h os.h \
  runtime.h fixnump.h align.h target-os.h darwin-os.h target-arch-os.h \
  arch-os-generic.inc target-arch.h globals.h core.h arch.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h genesis/vector.h interrupt.h \
  validate.h interr.h gc-internal.h genesis/code.h genesis/simple-fun.h \
  gc-assert.h code.h gencgc-internal.h hopscotch.h genesis/bignum.h \
  gc-private.h genesis/weak-pointer.h immobile-space.h pseudo-atomic.h \
  ../../output/build-id.inc genesis/gc-tables.h genesis/cons.h \
  genesis/hash-table.h genesis/layout.h var-io.h unaligned.h
