traceroot.o traceroot.d : traceroot.c sbcl.h genesis/config.h genesis/constants.h \
  arch.h os.h runtime.h fixnump.h align.h target-os.h darwin-os.h \
  target-arch-os.h arch-os-generic.inc target-arch.h thread.h globals.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h genesis/vector.h interrupt.h \
  validate.h lispregs.h target-lispregs.h genesis/regnames.h \
  gc-internal.h genesis/code.h genesis/simple-fun.h interr.h gc-assert.h \
  code.h gencgc-internal.h hopscotch.h genesis/bignum.h gc-private.h \
  genesis/weak-pointer.h immobile-space.h core.h genesis/closure.h \
  genesis/cons.h genesis/gc-tables.h genesis/hash-table.h \
  genesis/instance.h genesis/layout.h genesis/package.h pseudo-atomic.h \
  search.h genesis/avlnode.h genesis/sap.h
