immobile-space.o immobile-space.d : immobile-space.c gc.h sbcl.h genesis/config.h \
  genesis/constants.h os.h runtime.h fixnump.h align.h target-os.h \
  darwin-os.h target-arch-os.h arch-os-generic.inc target-arch.h \
  gc-internal.h genesis/code.h genesis/simple-fun.h thread.h globals.h \
  gencgc-alloc-region.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h genesis/vector.h interrupt.h \
  validate.h interr.h gc-assert.h code.h gencgc-internal.h hopscotch.h \
  genesis/bignum.h gc-private.h genesis/weak-pointer.h immobile-space.h \
  core.h genesis/gc-tables.h genesis/cons.h genesis/layout.h \
  forwarding-ptr.h pseudo-atomic.h var-io.h unaligned.h \
  genesis/hash-table.h genesis/package.h
