backtrace.o backtrace.d : backtrace.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h align.h globals.h os.h target-os.h darwin-os.h \
  target-arch-os.h arch-os-generic.inc target-arch.h interrupt.h \
  genesis/static-symbols.h thread.h gencgc-alloc-region.h gc.h \
  genesis/symbol.h genesis/thread.h genesis/fdefn.h genesis/vector.h \
  validate.h lispregs.h target-lispregs.h genesis/regnames.h arch.h \
  genesis/compiled-debug-fun.h genesis/compiled-debug-info.h \
  genesis/package.h genesis/primitive-objects.h genesis/array.h \
  genesis/bignum.h genesis/binding.h genesis/catch-block.h \
  genesis/closure.h genesis/code.h genesis/complex.h \
  genesis/complex-double-float.h genesis/complex-single-float.h \
  genesis/cons.h genesis/double-float.h genesis/funcallable-instance.h \
  genesis/instance.h genesis/ratio.h genesis/sap.h genesis/simd-pack.h \
  genesis/simd-pack-256.h genesis/simple-fun.h genesis/unwind-block.h \
  genesis/value-cell.h genesis/weak-pointer.h code.h gc-assert.h \
  interr.h var-io.h
