gencgc.o gencgc.d : gencgc.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h align.h os.h target-os.h darwin-os.h target-arch-os.h \
  arch-os-generic.inc target-arch.h interr.h globals.h interrupt.h \
  genesis/static-symbols.h thread.h gencgc-alloc-region.h gc.h \
  genesis/symbol.h genesis/thread.h genesis/fdefn.h genesis/vector.h \
  validate.h lispregs.h target-lispregs.h genesis/regnames.h arch.h \
  gc-internal.h genesis/code.h genesis/simple-fun.h gc-assert.h code.h \
  gencgc-internal.h hopscotch.h genesis/bignum.h gc-private.h \
  genesis/weak-pointer.h immobile-space.h core.h gencgc-private.h \
  pseudo-atomic.h alloc.h genesis/sap.h genesis/gc-tables.h save.h \
  genesis/hash-table.h genesis/instance.h genesis/layout.h \
  genesis/cons.h forwarding-ptr.h private-cons.inc
